"""

:copyright: Copyright 2013 by Luc Saffre.
:license: BSD, see LICENSE for more details.

"""
import os
import sys
import doctest
import unittest
import subprocess
from setuptools import find_packages

import six
        
class SubProcessTestCase(unittest.TestCase):
    "Deserves a docstring"
    project_root = NotImplementedError
    default_environ = dict()
    inheritable_envvars = ('VIRTUAL_ENV','PYTHONPATH','PATH')
    #~ maxDiff = None
    
    def run_packages_test(self,declared_packages):
        """
        Checks whether the `packages` parameter to setup seems correct.
        """
        found_packages = find_packages()
        found_packages.remove('tests') # if it exists, remove it
        found_packages.sort()
        declared_packages.sort()
        self.assertEqual(found_packages,declared_packages)
        
    def run_subprocess(self,args,**kw): 
        """
        Additional keywords can be 
        `cwd` : the working directory
        """
        env = dict(self.default_environ)
        for k in self.inheritable_envvars:
            v = os.environ.get(k,None)
            if v is not None:
                env[k] = v
        kw.update(env=env)
        #~ subprocess.check_output(args,**kw)
        #~ from StringIO import StringIO
        #~ buffer = StringIO()
        kw.update(stdout=subprocess.PIPE)
        kw.update(stderr=subprocess.STDOUT)
        p = subprocess.Popen(args,**kw)
        p.wait()
        rv = p.returncode
        #~ kw.update(stderr=buffer)
        #~ rv = subprocess.call(args,**kw)
        if rv != 0:
            cmd = ' '.join(args)
            #~ self.fail("%s returned %d:-----\n%s\n-----" % (cmd,rv,buffer.getvalue()))
            (out, err) = p.communicate()
            msg = "%s (%s) returned %d:\n-----\n%s\n-----" % (cmd,env,rv,out)
            print msg
            self.fail(msg)
        
    def run_simple_doctests(self,n,**kw): # env.simple_doctests
        #~ cmd = "python -m doctest %s" % filename    
        args = ["python"] 
        args += ["-m"]
        args += ["doctest"]
        args += [n]
        self.run_subprocess(args,**kw)
        
    def run_docs_django_tests(self,n,**kw): # django_doctests
        args = ["django-admin.py"] 
        args += ["test"]
        args += ["--settings=%s" % n]
        args += ["--failfast"]
        args += ["--verbosity=0"]
        args += ["--pythonpath=%s" % self.project_root.child('docs')]
        self.run_subprocess(args,**kw)

    def run_django_manage_test(self,db,**kw): # run_django_databases_tests
        p = self.project_root.child(*db.split('/'))
        args = ["python","manage.py"] 
        args += ["test"]
        #~ args += more
        args += ["--noinput"]
        args += ["--failfast"]
        #~ args += ["--settings=settings"]
        args += ["--pythonpath=%s" % p.absolute()]
        kw.update(cwd=p)
        self.run_subprocess(args,**kw)
        
    #~ def run_django_admin_tests(self,settings_module,**kw): # django_admin_tests
    def run_django_admin_test(self,settings_module,**kw): # django_admin_tests
        args = ["django-admin.py"] 
        args += ["test"]
        args += ["--settings=%s" % settings_module]
        args += ["--noinput"]
        args += ["--failfast"]
        args += ["--traceback"]
        args += ["--verbosity=0"]
        self.run_subprocess(args,**kw)
    
        #~ cmd = "django-admin.py test --settings=%s --verbosity=0 --failfast --traceback" % prj

    def run_docs_doctests(self,filename):
        """
        Run a simple doctest for specified file
        """
        filename = 'docs/' + filename
        #~ p = self.project_root.child(*filename.split('/')).parent
        #~ os.environ['DJANGO_SETTINGS_MODULE']='settings'
        #~ oldcwd = os.getcwd()
        #~ self.project_root.child('docs').chdir()
        #~ p.chdir()
        #~ sys.path.insert(0,'.')
        #~ print p
        sys.path.insert(0,'docs')
        import conf
        doctest.testfile(filename, module_relative=False,encoding='utf-8')
        del sys.path[0]
        #~ os.chdir(oldcwd)



"""
This module is being execfile'd from `setup.py`, `atelier/__init__.py`
and possibly some external tools, too.

:copyright: Copyright 2013 by Luc Saffre.
:license: BSD, see LICENSE for more details.
"""
SETUP_INFO = dict(
  name = 'atelier', 
  version = '0.0.1',
  install_requires = ['Sphinx','unipath','python_dateutil'],
  scripts = ['scripts/per_project'],
  description = "A collection of tools for software artists",
  license = 'Free BSD',
  test_suite = 'tests',
  author = 'Luc Saffre',
  author_email = 'luc.saffre@gmail.com',
  url = "http://code.google.com/p/python-atelier/",
  long_description="""\
`atelier` is my collection of tools for people who write and 
maintain multiple Python software projects.
It is not yet well documented, and so far there is 
nobody except me who uses it.
Let me know if you like it.
""",
  classifiers="""\
  Programming Language :: Python
  Programming Language :: Python :: 2.6
  Programming Language :: Python :: 2.7
  Development Status :: 4 - Beta
  Intended Audience :: Developers
  License :: OSI Approved :: BSD License
  Natural Language :: English
  Operating System :: OS Independent""".splitlines())

SETUP_INFO.update(packages = [str(n) for n in """
atelier
atelier.sphinxconf
""".splitlines() if n])
  



# -*- coding: utf-8 -*-
"""

This Sphinx extension defines the 
``blogger_year`` and 
``blogger_index``
directives and the ``blogref`` text role.


:copyright: Copyright 2011-2013 by Luc Saffre.
:license: BSD, see LICENSE for more details.

Thanks to 

- `Creating reStructuredText Directives 
  <http://docutils.sourceforge.net/docs/howto/rst-directives.html>`_


"""
  
import os
import sys
import calendar
import datetime
from StringIO import StringIO

from unipath import Path
#~ import lino

#~ from django.conf import settings

#~ from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned

from docutils import nodes, utils
from docutils import statemachine
from docutils.parsers.rst import directives
from docutils.parsers.rst import roles
from sphinx.util.compat import Directive

from atelier import rstgen

from atelier.sphinxconf import InsertInputDirective

import jinja2

templates = dict()
templates['calendar.rst'] = """
====
{{year}}
====

{{intro}}

.. |br| raw:: html

   <br />   

.. |sp| raw:: html

   <span style="color:white;">00</span>

{{calendar}}


"""

JINJA_ENV = jinja2.Environment(
    #~ extensions=['jinja2.ext.i18n'],
    loader=jinja2.DictLoader(templates))
    
    



class Year(object):
    """
    A :class:`Year` instance is created for each 
    `blogger_year` directive.
    """
    #~ _instances = dict()
    #~ def __init__(self,env,blogname,starting_year):
    #~ def __init__(self,env,blogname,year):
    def __init__(self,env):
        """
        :docname: the name of the document containing the `main_blogindex` directive
        :starting_year: all years before this year will be pruned
        """
        
        blogname, year, index = env.docname.rsplit('/',3)
        if index != 'index':
            raise Exception("Allowed only in /<blogname>/<year>/index.rst files")
        self.year = int(year)
        
        #~ print "20130113 Year.__init__", blogname, self.year
        #~ self.blogname = blogname
        self.days = set()
        #~ self.years = set()
        #~ self.starting_year = int(starting_year)
        top = os.path.dirname(env.doc2path(env.docname))
        #~ print top
        for (dirpath, dirnames, filenames) in os.walk(top):
            del dirnames[:] # don't descend another level
            #~ unused, year = dirpath.rsplit(os.path.sep,2)
            #~ year = int(year)
            #~ assert year in self.years
            for fn in filenames:
                if len(fn) == 8 and fn.endswith('.rst'):
                    d = docname_to_day(self.year,fn[:-4])
                    self.days.add(d)
                    #~ self.years.add(s)
                        
        #~ self.years = sorted(self.years)
        if not hasattr(env,'blog_instances'):
            env.blog_instances = dict()
        years = env.blog_instances.setdefault(blogname,dict())
        years[self.year] = self
                        
        
        
"""
docs/conf.py
docs/blog/index.rst --> contains a main_blogindex directive (hidden toctree)
docs/blog/2013/index.rst --> contains a blogger_year directive (calendar)
docs/blog/2013/0107.rst --> a blog entry
docs/blog/2010/0107.rst

"""
        
   
    
class MainBlogIndexDirective(InsertInputDirective):
    """
    Directive to insert a blog master archive page toctree
    """
    #~ required_arguments = 1
    #~ allow_titles = True
    raw_insert = True
  
    def get_rst(self):
        #~ print 'MainBlogIndexDirective.get_rst()'
        env = self.state.document.settings.env
        intro = '\n'.join(self.content)
        #~ dn  = os.path.dirname(env.doc2path(env.docname))
        #~ year = os.path.split(dn)[-1]
        blogname, index = env.docname.rsplit('/',2)
        if index != 'index':
            raise Exception("Allowed only inside index.rst file")
        #~ blog = Blog.get_or_create(env,blogname,self.arguments[0])
        text = intro
        text += """

.. toctree::
    :maxdepth: 2

"""
        years = list(env.blog_instances.get(blogname).values())
        def f(a,b): 
            return cmp(a.year,b.year)
        years.sort(f)
        for blogger_year in years:
        #~ for year in blog.years:
            text += """
    %d/index""" % blogger_year.year
  
        text += "\n"
        #~ print text
        return text
   
   
   

class YearBlogIndexDirective(InsertInputDirective):
    """
    Directive to insert a year's calendar
    """
    #~ required_arguments = 1
    #~ allow_titles = True
    raw_insert = True
    
      
    def get_rst(self):
        from djangosite.dbutils import monthname
  
        #~ year = self.arguments[0]
        env = self.state.document.settings.env
        today = datetime.date.today()
        
        #~ dn  = os.path.dirname(env.doc2path(env.docname))
        #~ year = os.path.split(dn)[-1]
        blogger_year = Year(env)
        #~ blog = Blog.get_or_create(env,blogname)
        
        tpl = JINJA_ENV.get_template('calendar.rst')
        
        intro = '\n'.join(self.content)
        cal = calendar.Calendar()
        text = ''
        
        for month in range(1,13):
            
            text += """        
            
.. |M%02d| replace::  **%s**""" % (month,monthname(month))
            
            weeknum = None
            #~ text += "\n  |br| Mo Tu We Th Fr Sa Su "
            for day in cal.itermonthdates(blogger_year.year,month):
                iso_year,iso_week,iso_day = day.isocalendar()
                if iso_week != weeknum:
                    text += "\n  |br|"
                    weeknum = iso_week
                if day.month == month:
                    label = "%02d" % day.day
                    docname = "%02d%02d" % (day.month,day.day)
                    if blogger_year.year == iso_year and day in blogger_year.days:
                        text += " :doc:`%s <%s>` " % (label,docname)
                    elif day > today:
                        text += ' |sp| '
                    else:
                        text += ' ' + label + ' '
                else:
                    text += ' |sp| '
                
            
            
        text += """
        
===== ===== =====
|M01| |M02| |M03|
|M04| |M05| |M06|
|M07| |M08| |M09|
|M10| |M11| |M12|
===== ===== =====
        
        """
        
        text += """

.. toctree::
    :hidden:
    :maxdepth: 2
    
"""
        
        days = sorted(blogger_year.days)
        for day in days:
            text += """
    %02d%02d""" % (day.month,day.day)
        
        return tpl.render(
            calendar=text,
            intro=intro,
            year=blogger_year.year,
            days=blogger_year.days)
        

def docname_to_day(year,s):
    #~ print fn
    month = int(s[:2])
    day = int(s[2:])
    return datetime.date(year,month,day)
  
  
#~ class ChangedDirective(InsertInputDirective):
  
    #~ def get_rst(self):
        #~ env = self.state.document.settings.env
        #~ blogname, year, monthday = env.docname.rsplit('/',3)
        #~ # raise Exception("Allowed only in blog entries")
        
        #~ year = int(year)
        #~ day = docname_to_day(year,monthname)
        
        #~ if not hasattr(env,'changed_items'):
            #~ env.changed_items = dict()
        #~ env.changed_items
        #~ for item in self.content:
            #~ entries = env.changed_items.setdefault(item,dict())
            #~ entries.setdefault(env.docname)
            
            

def setup(app):
    #~ app.add_node(blogindex)
    #~ app.add_node(blogindex,html=(visit_blogindex,depart_blogindex))
    #~ app.add_directive('changed', ChangedDirective)
    app.add_directive('blogger_year', YearBlogIndexDirective)
    app.add_directive('blogger_index', MainBlogIndexDirective)
    





# -*- coding: UTF-8 -*-
"""

:copyright: Copyright 2009-2013 by Luc Saffre.
:license: BSD, see LICENSE for more details.
"""

from __future__ import unicode_literals

import sys
import types
import datetime
from dateutil import parser as dateparser



class AttrDict(dict):
    """
    Dictionary-like helper object.
    
    Usage example:
    
    >>> from atelier.utils import AttrDict
    >>> a = AttrDict()
    >>> a.define('foo',1)
    >>> a.define('bar','baz',2)
    >>> print a
    {u'foo': 1, u'bar': {u'baz': 2}}
    >>> print a.foo
    1
    >>> print a.bar.baz
    2
    >>> print a.resolve('bar.baz')
    2
    >>> print a.bar
    {u'baz': 2}
    
    """
  
    #~ def __getattr__(self, name):
        #~ return self[name]
        
    def __getattr__(self, name,*args):
        return self.get(name,*args)
        #~ raise AttributeError("%r has no attribute '%s'" % (self,name))
        
    def define2(self,name,value):
        return self.define(*name.split('.')+[value])
        
    def define(self,*args):
        "args must be a series of names followed by the value"
        assert len(args) >= 2
        d = s = self
        for n in args[:-2]:
            d = s.get(n,None)
            if d is None:
                d = AttrDict()
                s[n] = d
            s = d
        oldvalue = d.get(args[-2],None)
        #~ if oldvalue is not None:
            #~ print 20120217, "Overriding %s from %r to %r" % (
              #~ '.'.join(args[:-1]),
              #~ oldvalue,
              #~ args[-1]
              #~ )
        d[args[-2]] = args[-1]
        return oldvalue
    
    def resolve(self,name,default=None):
        """
        return an attribute with dotted name
        """
        o = self
        for part in name.split('.'):
            o = getattr(o,part,default)
            # o = o.__getattr__(part)
        return o


def iif(condition,true_value,false_value): 
    """
    "Inline If" : an ``if`` statement as a function.
    
    Examples:
    
    >>> import six
    >>> from atelier.utils import iif
    >>> six.print_("Hello, %s world!" % iif(1+1==2,"real","imaginary"))
    Hello, real world!
    
    """
    if condition: return true_value
    return false_value
    
def i2d(i):
    """
    Convert `int` to `date`. Examples:
    
    >>> i2d(20121224)
    datetime.date(2012, 12, 24)
    
    """
    d = dateparser.parse(str(i))
    d = datetime.date(d.year,d.month,d.day)
    #print i, "->", v
    return d
    
def i2t(s):
    """
    Convert `int` to `time`. Examples:
    
    >>> i2t(815)
    datetime.time(8, 15)
    
    >>> i2t(1230)
    datetime.time(12, 30)
    
    >>> i2t(12)
    datetime.time(12, 0)
    
    >>> i2t(1)
    datetime.time(1, 0)
    
    """
    s = str(s)
    if len(s) == 4:
        return datetime.time(int(s[:2]),int(s[2:]))
    if len(s) == 3:
        return datetime.time(int(s[:1]),int(s[1:]))
    if len(s) <= 2:
        return datetime.time(int(s),0)
    raise ValueError(s)

    
    
    

def ispure(s):
    """
    Returns `True` if the specified string `s` is either a unicode 
    string or contains only ASCII characters.
    """
    if s is None: return True 
    if type(s) == types.UnicodeType:
        return True
    if type(s) == types.StringType:
        try:
            s.decode('ascii')
        except UnicodeDecodeError as e:
            return False
        return True
    return False

def assert_pure(s):
    #~ assert ispure(s), "%r: not pure" % s
    if s is None: return 
    if isinstance(s,unicode):
        return True
    try:
        s.decode('ascii')
    except UnicodeDecodeError as e:
        raise Exception("%r is not pure : %s" % (s,e))
     



def confirm(prompt=None):
    """
    Ask for user confirmation from the console.
    """
    while True:
        ln = raw_input(prompt)
        if ln.lower() in ('y','j','o'):
            return True
        if ln.lower() == 'n':
            return False
        print "Please anwer Y or N"



def indentation(s):
    r"""
    Examples:
    
    >>> from atelier.utils import indentation
    >>> indentation("")
    0
    >>> indentation("foo")
    0
    >>> indentation(" foo")
    1
    
    """
    return len(s)-len(s.lstrip())


def unindent(s):
    r"""
    Reduces indentation of a docstring to the minimum.
    Empty lines don't count.
    
    Examples:
    
    >>> from atelier.utils import unindent
    >>> unindent('')
    u''
    >>> print unindent('''
    ...   foo
    ...     foo
    ... ''')
    <BLANKLINE>
    foo
      foo
    >>> print unindent('''
    ... foo
    ...     foo
    ... ''')
    <BLANKLINE>
    foo
        foo
    """
    lines = s.splitlines()
    if len(lines) == 0: 
        return s.lstrip()
    mini = sys.maxint
    for ln in lines:
        ln = ln.rstrip()
        if len(ln) > 0: 
            mini = min(mini,indentation(ln))
            if mini == 0:
                break
    if mini == sys.maxint: 
        return s
    return '\n'.join([ln[mini:] for ln in lines])
    



# -*- coding: utf-8 -*-
"""

Sphinx setup used to build the Lino documentation.

:copyright: Copyright 2011-2013 by Luc Saffre.
:license: BSD, see LICENSE for more details.

Thanks to 

- `Creating reStructuredText Directives 
  <http://docutils.sourceforge.net/docs/howto/rst-directives.html>`_


"""
  
import os
import sys
import calendar
import datetime
from StringIO import StringIO

from unipath import Path
#~ import lino

#~ from django.conf import settings

#~ from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned

from docutils import nodes, utils
from docutils import statemachine
from docutils.parsers.rst import directives
from docutils.parsers.rst import roles
from sphinx.util.compat import Directive
from sphinx.util.nodes import nested_parse_with_titles
from sphinx.util.nodes import split_explicit_title

from atelier import rstgen

from atelier.utils import i2d


#~ class ScreenshotDirective(directives.images.Image):
    #~ """
    #~ Directive to insert a screenshot.
    #~ """
    #~ def run(self):
        #~ assert len(self.arguments) == 1
        #~ # name = '/../gen/screenshots/' + self.arguments[0]
        #~ name = '/gen/screenshots/' + self.arguments[0]
        #~ self.arguments = [name]
        #~ (image_node,) = directives.images.Image.run(self)
        #~ return [image_node]





def srcref(mod):
    """
    Return the `source file name` for usage by Sphinx's ``srcref`` role.
    Returns None if the source file is empty (which happens e.g. for __init__.py 
    files whose only purpose is to mark a package).
    
    >>> from atelier.sphinxconf import srcref
    >>> from lino.utils import log
    >>> print srcref(log)
    lino/utils/log.py

    >>> from lino import utils
    >>> print srcref(utils)
    lino/utils/__init__.py
    
    >>> from lino.management import commands
    >>> print srcref(commands)
    None

    >>> from lino_welfare.demo import settings
    >>> print srcref(settings)
    lino_welfare/demo/settings.py

    """
    root_module_name = mod.__name__.split('.')[0]
    root_mod = __import__(root_module_name)
    #~ if not mod.__name__.startswith('lino.'): 
        #~ return
    srcref = mod.__file__
    if srcref.endswith('.pyc'):
        srcref = srcref[:-1]
    if os.stat(srcref).st_size == 0:
        return 
    #~ srcref = srcref[len(lino.__file__)-17:]
    srcref = srcref[len(Path(root_mod.__file__).ancestor(2))+1:]
    srcref = srcref.replace(os.path.sep,'/')
    return srcref
    


def autodoc_skip_member(app, what, name, obj, skip, options):
    if name != '__builtins__':
        #~ print 'autodoc_skip_member', what, repr(name), repr(obj)
    
        if what == 'class':
            if name.endswith('MultipleObjectsReturned'):
                return True
            if name.endswith('DoesNotExist'):
                return True
                
            #~ if isinstance(obj,ObjectDoesNotExist) \
              #~ or isinstance(obj,MultipleObjectsReturned): 
                #~ return True
        
    #~ if what == 'exception': 
        #~ print 'autodoc_skip_member',what, repr(name), repr(obj), skip
        #~ return True
        

#~ SIDEBAR = """
#~ (This module's source code is available at :srcref:`/%s`)

#~ """  
#~ SIDEBAR = """
#~ (source code: :srcref:`/%s`)

#~ """  

SIDEBAR = """
(:srcref:`source code </%s>`)

"""  

#~ SIDEBAR = """
#~ .. sidebar:: Use the source, Luke

  #~ We generally recommend to also consult the source code.
  #~ This module's source code is available at
  #~ :srcref:`/%s.py`

#~ """  


    
def autodoc_add_srcref(app,what,name,obj,options,lines):
    if what == 'module':
        s = srcref(obj)
        if s:
            #~ srcref = name.replace('.','/')
            s = (SIDEBAR % s).splitlines()
            s.reverse()
            for ln in s:
                lines.insert(0,ln)
            #~ lines.insert(0,'')
            #~ lines.insert(0,'(We also recommend to read the source code at :srcref:`/%s.py`)' % name.replace('.','/'))
    


    
    


class InsertInputDirective(Directive):
    """
    Base class for directives that work by generating rst markup
    to be forwarded to `state_machine.insert_input()`.
    """
    titles_allowed = False
    has_content = True
    debug = False
    raw_insert = False
    
    def get_rst(self):
        raise NotImplementedErrro()
        
    #~ def run(self):
        #~ out = self.get_rst()
        #~ env = self.state.document.settings.env
        #~ if self.debug:
            #~ print env.docname
            #~ print '-' * 50
            #~ print out
            #~ print '-' * 50
        #~ self.state_machine.insert_input(out.splitlines(),out)
        #~ return []
        
    #~ class InsertTableDirective(InsertInputDirective):
    
    def run(self):
        self.env = self.state.document.settings.env
        output = self.get_rst()
        #~ output = output.decode('utf-8')
        
        if self.debug:
            print self.state.document.settings.env.docname
            print '-' * 50
            print output
            print '-' * 50
        
        content = statemachine.StringList(output.splitlines())
        
        if self.raw_insert:
          
            self.state_machine.insert_input(content,output)
            return []
            
        
        if self.titles_allowed:
            node = nodes.section()
            # necessary so that the child nodes get the right source/line set
            node.document = self.state.document
            nested_parse_with_titles(self.state, content, node)
        else:
            node = nodes.paragraph()
            node.document = self.state.document
            self.state.nested_parse(content, self.content_offset, node)


        
        # following lines originally copied from 
        # docutils.parsers.rst.directives.tables.RSTTable 
        #~ title, messages = self.make_title()
        #~ node = nodes.Element()          # anonymous container for parsing
        #~ self.state.nested_parse(content, self.content_offset, node)
        #~ if len(node) != 1 or not isinstance(node[0], nodes.table):
            #~ error = self.state_machine.reporter.error(
                #~ 'Error parsing content block for the "%s" directive: exactly '
                #~ 'one table expected.' % self.name, nodes.literal_block(
                #~ self.block_text, self.block_text), line=self.lineno)
            #~ return [error]
        #~ return [x for x in node]
        return list(node)
        
        #~ table_node = node[0]
        #~ table_node['classes'] += self.options.get('class', [])
        #~ return [table_node] 


    

    
class Py2rstDirective(InsertInputDirective):
    """
    Run a Python code block and interpret the output as if it 
    were rst source.
    """
    titles_allowed = True
    has_content = True
    debug = False
    #~ def get_rst(self):
        #~ return self.output_from_exec('\n'.join(self.content))
        
    def get_context(self): 
        context = dict()
        context.update(self=self)
        context.update(env=self.state.document.settings.env)
        return context
    
    def get_rst(self):
        if not self.content:
            warning = self.state_machine.reporter.warning(
                'Content block expected for the "%s" directive; none found.'
                % self.name, nodes.literal_block(
                self.block_text, self.block_text), line=self.lineno)
            return [warning]
            
        #~ raise Exception("20130331 %r" % self.content)
        code = '\n'.join(self.content)
        return self.output_from_exec(code)        
        
    def output_from_exec(self,code):
        old = sys.stdout
        buffer = StringIO()
        sys.stdout = buffer
        context = self.get_context()
        exec(code,context)
        sys.stdout = old
        s = buffer.getvalue()
        #~ print 20130331, type(s)
        return s
        


class Django2rstDirective(Py2rstDirective):
    def get_context(self):
        from djangosite.dbutils import set_language
        from django.conf import settings
        context = super(Django2rstDirective,self).get_context()
        lng = self.state.document.settings.env.config.language
        set_language(lng)
        context.update(settings=settings)
        context.update(settings.SITE.modules)
        return context
        
        
        
#~ class DjangoTableDirective(InsertInputDirective):
    #~ def get_rst(self):
        #~ assert len(self.content) == 1
        #~ code = '\n'.join(self.content)
        
        #~ from django.conf import settings    
        #~ print .jobs.Candidatures.request(limit=5).to_rst()
        
        #~ code = """
        #~ """
        #~ old = sys.stdout
        #~ buffer = StringIO()
        #~ sys.stdout = buffer
        #~ context = dict()
        #~ context.update(settings.SITE.modules)
        #~ context = dict(settings=settings)
        #~ exec(code,context)
        #~ sys.stdout = old
        #~ return buffer.getvalue()
        
class TextImageDirective(InsertInputDirective):
    """
    See Blog entry 2013/0116 for documentation.
    """
    required_arguments = 1
    final_argument_whitespace = True
    option_spec = dict(scale=directives.unchanged)
    #~ optional_arguments = 4
    
    def get_rst(self):
        #~ print 'MainBlogIndexDirective.get_rst()'
        #~ env = self.state.document.settings.env
        #~ print self.arguments, self.options, self.content
        left = '\n'.join(self.content)
        right = ''
        for arg in self.arguments[0].split():
            right += '.. figure:: %s\n' % arg
            for i in self.options.items():
                right += "  :%s: %s\n" % i
            right += "\n  %s\n\n" % arg
            #~ right += "\n  \n\n" % arg
            
        return rstgen.table(["",""],[[left,right]],show_headers=False)
    
class ComplexTableDirective(InsertInputDirective):
    """
    The `complextable` directive is used to create tables
    with complex cell content
    
    Usage example::
    
      .. complextable::
    
        A1
        <NEXTCELL>
        A2
        <NEXTROW>
        B1
        <NEXTCELL>
        B2
        
    
    Result:
    
    .. complextable::
    
        A1
        <NEXTCELL>
        A2
        <NEXTROW>
        B1
        <NEXTCELL>
        B2
        
        
    See Blog entry 2013/0116 for documentation.
    """
    required_arguments = 0
    final_argument_whitespace = True
    option_spec = dict(header=directives.flag)
    #~ option_spec = dict(scale=unchanged)
    #~ optional_arguments = 4
    
    def get_rst(self):
        #~ print 'MainBlogIndexDirective.get_rst()'
        #~ env = self.state.document.settings.env
        #~ print self.arguments, self.options, self.content
        cellsep = '<NEXTCELL>'
        rowsep = '<NEXTROW>'
        if len(self.arguments) > 0:
            cellsep = self.arguments[0]
        if len(self.arguments) > 1:
            rowsep = self.arguments[1]
        
        content = '\n'.join(self.content)
        rows = []
        colcount = None
        
        for row in content.split(rowsep):
            cells = [cell.strip() for cell in row.split(cellsep)]
            if colcount is None:
                colcount = len(cells)
            else:
                assert colcount == len(cells)
            rows.append(cells)
              
        if 'header' in self.options:
            return rstgen.table(rows[0],rows[1:])
            
        return rstgen.table([""] * colcount,rows,show_headers=False)




def get_blog_url(today):
    blogger_project = "lino"
    url_root = "http://code.google.com/p/%s/source/browse/" % blogger_project
    parts = ('docs','blog',str(today.year),today.strftime("%m%d.rst"))
    url = url_root + "/".join(parts)
    return url


def blogref_role(name, rawtext, text, lineno, inliner,options={}, content=[]):
    """
    Inserts a reference to the blog entry of the specified date.
    
    Instead of writing ``:doc:`/blog/2011/0406```
    it is better to write ``:blogref:`20110406```
    because the latter works between Sphinx trees and also supports archived blog entries.
    
    """
    # thanks to http://docutils.sourceforge.net/docs/howto/rst-roles.html
    # this code originally from roles.pep_reference_role
    #~ print 20130315, rawtext, text, utils.unescape(text)
    has_explicit_title, title, target = split_explicit_title(text)
    try:
        date = i2d(int(target))
    except ValueError:
        msg = inliner.reporter.error(
            'Invalid text %r: must be an integer date of style "20130315" .'
            % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    #~ print repr(env)
    #~ raise Exception(20130315)
    #~ ref = inliner.document.settings.pep_base_url
           #~ + inliner.document.settings.pep_file_url_template % date)
    roles.set_classes(options)
    #~ from django.conf import settings
    #~ shown_text = settings.SITE.dtos(date)
    env = inliner.document.settings.env        
    if not has_explicit_title:
        title = date.strftime(env.settings.get('today_fmt','%Y-%m-%d'))
    title = utils.unescape(title)
    return [nodes.reference(rawtext, title, 
                            refuri=get_blog_url(date),
                            **options)], []
    
    









#~ def configure(filename,globals_dict,settings_module_name='settings'):
def configure(globals_dict,settings_module_name=None):
    """
    To be callsed from inside the Sphinx `conf.py` as follows::
    
      from djangosite.utils.sphinxconf import configure
      configure(globals())

    This contains the things that all my Sphinx docs configuration 
    files have in common.
    
    """
    filename = globals_dict.get('__file__')
    DOCSDIR = Path(filename).parent.absolute()
    sys.path.append(DOCSDIR)

    HGWORK = DOCSDIR.ancestor(2)
    intersphinx_mapping = dict()
    for n in ('atelier','site','north','lino','welfare','garden'):
        p = Path(HGWORK,n,'docs','.build','objects.inv')
        if p.exists():
            intersphinx_mapping[n] = ('http://%s.lino-framework.org' % n,p)
    p = Path(HGWORK,'welfare','userdocs','.build','fr','objects.inv')
    if p.exists():
        intersphinx_mapping[n] = ('http://welfare-user.lino-framework.org',p)
    #~ else:
        #~ raise Exception("%s does not exist" % p)
    #~ intersphinx_mapping.update(django = (
        #~ 'http://docs.djangoproject.com/en/dev/', 
        #~ 'http://docs.djangoproject.com/en/dev/_objects/'))
    globals_dict.update(intersphinx_mapping=intersphinx_mapping)
    
    globals_dict.update(extensions = [
      'sphinx.ext.autodoc',
      #~ 'sphinx.ext.autosummary',
      'sphinx.ext.inheritance_diagram',
      'sphinx.ext.todo',
      'sphinx.ext.extlinks',
      'sphinx.ext.graphviz',
      'sphinx.ext.intersphinx',
      'sphinxcontrib.newsfeed', # no i18n, no discovery, only one entry per doc, 
      #~ 'sphinx.ext.doctest',
    ])
    
    if settings_module_name is not None:
        #~ os.environ['DJANGO_SETTINGS_MODULE'] = 'north.docs_settings'
        os.environ['DJANGO_SETTINGS_MODULE'] = settings_module_name
        """
        Trigger loading of Djangos model cache in order to avoid side effects that 
        would occur when this happens later while importing one of the models modules.
        """
        from django.conf import settings
        #~ settings.SITE # must at least access some variable in the settings
        settings.SITE.startup()
    globals_dict.update(setup=setup)

        

def setup2(app):
    # also used by `vor/conf.py`
    app.add_directive('complextable', ComplexTableDirective)
    app.add_directive('py2rst', Py2rstDirective)
    app.add_directive('django2rst', Django2rstDirective)
    #~ app.add_directive('linotable', InsertTableDirective)
    
def setup(app):
    """
    The Sphinx setup function used for Lino-related documentation trees.
   
    """
    app.add_object_type(directivename='xfile',rolename='xfile',
      indextemplate='pair: %s; file')
    app.add_object_type(directivename='setting',rolename='setting',
      indextemplate='pair: %s; setting')
    #~ app.add_object_type(directivename='model',rolename='model',
      #~ indextemplate='pair: %s; model')
    #~ app.add_object_type(directivename='field',rolename='field',
      #~ indextemplate='pair: %s; field')
    app.add_object_type(directivename='screenshot',rolename='screen',
      indextemplate='pair: %s; screenshot')
    app.add_object_type(directivename='modattr',rolename='modattr',
      indextemplate='pair: %s; model attribute')
    app.add_object_type(directivename='model',rolename='model',indextemplate='pair: %s; model')
    #app.connect('build-finished', handle_finished)
    
    app.connect('autodoc-skip-member',autodoc_skip_member)
    app.connect('autodoc-process-docstring', autodoc_add_srcref)
    
    app.add_directive('textimage', TextImageDirective)

    roles.register_canonical_role('blogref', blogref_role)
    
    setup2(app)
    #~ app.add_directive('screenshot', ScreenshotDirective)
    #~ app.add_config_value('screenshots_root', '/screenshots/', 'html')

    #~ from djangosite.utils import doctest
    #~ doctest.setup(app)
    
#~ print "OK"    


def version2rst(self,m):
    """
    used in docs/released/index.rst
    """
    v = m.__version__
    if v.endswith('+'):
        v = v[:-1]
        print "The current stable release is :doc:`%s`." % v 
        print "We are working on a future version in the code repository."
    elif v.endswith('pre'):
        print "We're currently working on :doc:`%s`." % v[:-3]
    else:
        print "The current stable release is :doc:`%s`." % v 
        #~ print "We're currently working on :doc:`coming`."






# -*- coding: UTF-8 -*-
"""
The fabric tasks I use to manage my projects.
Use at your own risk.

To be used by creating a `fabfile.py` with at least the following 
two lines::

  from atelier.fablib import *
  setup_from_project("foobar")
  
  
  env.django_admin_tests.append(...)
  env.simple_doctests.append(...)
  env.tolerate_sphinx_warnings = True
  env.demo_databases.append('foobar.demo.settings')
  
Where "foobar" is the name of your main package.
  
This fablib uses the following `env` keys:

- `tolerate_sphinx_warnings` : whether `sphinx-build html` should tolerate warnings.

- (consult the source code)

:copyright: Copyright 2013 by Luc Saffre.
:license: BSD, see LICENSE for more details.
"""
import os
import sys
import doctest
import pkg_resources
import textwrap

#~ def clean_sys_path():
    #~ # print sys.path
    #~ if sys.path[0] == '':
        #~ del sys.path[0]
        #~ print "Deleted working directory from PYTHONPATH"


import datetime
import unittest
#~ import subprocess
from setuptools import find_packages
from unipath import Path
import sphinx
#~ import six
#~ from distutils.core import run_setup

#~ import djangosite ; print djangosite.__file__
from atelier.utils import AttrDict
from atelier import rstgen
#~ from timtools.tools.synchronizer import Synchronizer


from fabric.api import env, local, task, prompt
from fabric.utils import abort, fastprint, puts, warn
from fabric.contrib.console import confirm
from fabric.api import lcd

#~ LONG_DATE_FORMAT = 

PROJECTS = []

class RstFile(object):
    def __init__(self,local_root,url_root,parts):
        self.path = local_root.child(*parts)
        self.url = url_root + "/".join(parts)
        #~ self.parts = parts
        
      

class Project(object):
    def __init__(self,i,name):
        self.index = i
        self.name = name
        self.dist = pkg_resources.get_distribution(name)
        self.module = __import__(name)
    
def load_projects():
    if len(PROJECTS) > 0: 
        return 
    for i,prj in enumerate(env.projects.split()):
        PROJECTS.append(Project(i,prj))
      

def setup_from_project(main_package=None):
  
    #~ env.docs_rsync_dest = 'luc@lino-framework.org'
    #~ env.sdist_dir = '../lino/docs/dl'

    #~ HOME = Path(os.path.expanduser("~"))
    #~ REMOTE = AttrDict(
      #~ lf='lino-framework.org')
      
    env.ROOTDIR = Path().absolute()

    env.project_name = env.ROOTDIR.absolute().name

    env.setdefault('long_date_format',"%Y%m%d (%A, %d %B %Y)"  )
    #~ env.setdefault('work_root','')
    env.work_root = Path(env.work_root)
    env.sdist_dir = Path(env.sdist_dir)
    env.django_doctests = []
    env.django_admin_tests = []
    env.bash_tests = []
    #~ env.django_databases = []
    env.simple_doctests = []
    env.docs_doctests = []
    env.main_package = main_package
    env.tolerate_sphinx_warnings = False
    env.demo_databases = []
    

    env.languages = None
    
    #~ print env.project_name

    env.DOCSDIR = Path(env.ROOTDIR,'docs')
    #~ env.BUILDDIR = Path(env.DOCSDIR,'.build')

    if not Path(env.ROOTDIR,'setup.py').exists():
        raise Exception("You must call 'fab' from a project's root directory.")
        
    if env.main_package:
        execfile(env.ROOTDIR.child(env.main_package,'setup_info.py'),globals()) # will set SETUP_INFO 
        env.SETUP_INFO = SETUP_INFO
    else:
        env.SETUP_INFO = None
        
    


#~ def confirm(msg,default='y',others='n',**override_callbacks):
    #~ text = "%s [%s%s]" % (msg,default.upper(),others)
    #~ def y(): return True
    #~ # def n(): abort("Missing user confirmation for:\n%s" % msg)
    #~ def n(): abort("Missing user confirmation")
    #~ callbacks = dict(y=y,n=n)
    #~ callbacks.update(override_callbacks)
    #~ while True:
        #~ answer = prompt(text)
        #~ # answer = raw_input(prompt)
        #~ if not answer: 
            #~ answer = default
        #~ answer = answer.lower()
        #~ if answer: 
            #~ return callbacks.get(answer)()
            
def must_confirm(*args,**kw):
    if not confirm(*args,**kw):
        abort("Dann eben nicht...")
        
def must_exist(p):
    if not p.exists():
        abort("No such file: %s" % p.absolute())
        
def rmtree_after_confirm(p):
    if not p.exists(): return
    must_confirm("OK to remove %s and everything under it?" % p.absolute())
    p.rmtree()
    


def get_locale_dir():
    if not env.main_package:
        return None # abort("No main_package")
    p = env.ROOTDIR.child(env.main_package,"locale")
    if not p.isdir():
        return None # abort("Directory %s does not exist." % p)
    return p

   
@task(alias='mm')
def make_messages():
    "Extract messages, then initialize and update all catalogs"
    extract_messages()
    init_catalog_code()
    update_catalog_code()
    
    extract_messages_userdocs()
    setup_babel_userdocs('init_catalog')
    setup_babel_userdocs('update_catalog')
    
#~ @task(alias='em')
def extract_messages():
    """Extract messages from source files to .pot file"""
    locale_dir = get_locale_dir()
    if locale_dir is None: return 
    args = ["python", "setup.py"]
    args += [ "extract_messages"]
    args += [ "-o", locale_dir.child("django.pot")]
    cmd = ' '.join(args)
    #~ must_confirm(cmd)
    local(cmd)
    
#~ @task(alias='emu')
def extract_messages_userdocs(): 
    """
    Run the Sphinx gettext builder on userdocs.
    """
    args = ['sphinx-build','-b','gettext']
    userdocs = env.ROOTDIR.child('userdocs')
    if not userdocs.isdir():
        return # abort("Directory %s does not exist." % userdocs)
    #~ args += cmdline_args
    #~ args += ['-a'] # all files, not only outdated
    #~ args += ['-P'] # no postmortem
    #~ args += ['-Q'] # no output
    if not env.tolerate_sphinx_warnings:
        args += ['-W'] # consider warnings as errors
    #~ args += ['-w',env.DOCSDIR.child('warnings.txt')]
    args += [userdocs]
    args += [userdocs.child("translations")]
    cmd = ' '.join(args)
    local(cmd)
    
    

@task(alias='rename')
def rename_data_url_friendly():
    data_dir = env.ROOTDIR.child('docs','data')
    #~ print list(data_dir.listdir(names_only=True))
    print list(data_dir.walk())
    
def setup_babel_userdocs(babelcmd):
    """Create userdocs .po files if necessary."""
    userdocs = env.ROOTDIR.child('userdocs')
    if not userdocs.isdir(): return
    locale_dir = userdocs.child('translations')
    for domain in locale_dir.listdir('*.pot',names_only=True):
        domain = domain[:-4]
        for loc in env.languages:
            po_file = Path(locale_dir,loc,'LC_MESSAGES','%s.po' % domain)
            mo_file = Path(locale_dir,loc,'LC_MESSAGES','%s.mo' % domain)
            pot_file = Path(locale_dir,'%s.pot' % domain)
            if babelcmd == 'init_catalog' and po_file.exists():
                print "Skip %s because file exists." % po_file
            #~ elif babelcmd == 'compile_catalog' and not mo_file.needs_update(po_file):
                #~ print "Skip %s because newer than .po" % mo_file
            else:
                args = ["python", "setup.py"]
                args += [ babelcmd ]
                args += [ "-l" , loc ]
                args += [ "--domain", domain]
                args += [ "-d" , locale_dir ]
                #~ args += [ "-o" , po_file ]
                #~ if babelcmd == 'init_catalog':
                if babelcmd == 'compile_catalog':
                    args += [ "-i" , po_file ]
                else:
                    args += [ "-i" , pot_file ]
                cmd = ' '.join(args)
                #~ must_confirm(cmd)
                local(cmd)

#~ @task(alias='imu')
#~ def init_catalog_userdocs():
    #~ setup_babel_userdocs('init_catalog')

#~ @task(alias='umu')
#~ def update_catalog_userdocs():
    #~ setup_babel_userdocs('update_catalog')

@task(alias='cmu')
def compile_catalog_userdocs():
    setup_babel_userdocs('compile_catalog')


#~ @task(alias='im')
def init_catalog_code():
    """Create code .po files if necessary."""
    locale_dir = get_locale_dir()
    if locale_dir is None: return 
    for loc in env.languages:
        f = locale_dir.child(loc,'LC_MESSAGES','django.po')
        if f.exists():
            print "Skip %s because file exists." % f
        else:
            args = ["python", "setup.py"]
            args += [ "init_catalog"]
            args += [ "--domain django"]
            args += [ "-l" , loc ]
            args += [ "-d" , locale_dir ]
            #~ args += [ "-o" , f ]
            args += [ "-i" , locale_dir.child('django.pot') ]
            cmd = ' '.join(args)
            must_confirm(cmd)
            local(cmd)



#~ @task(alias='um')
def update_catalog_code():
    """Update .po files from .pot file."""
    locale_dir = get_locale_dir()
    if locale_dir is None: return 
    for loc in env.languages:
        args = ["python", "setup.py"]
        args += [ "update_catalog"]
        args += [ "--domain django"]
        args += [ "-d" , locale_dir ]
        args += [ "-i", locale_dir.child("django.pot")]
        args += [ "-l" , loc ]
        cmd = ' '.join(args)
        #~ must_confirm(cmd)
        local(cmd)

@task(alias='cm')
def compile_catalog():
    """Compile .po files to .mo files."""
    locale_dir = get_locale_dir()
    if locale_dir is None: return 
    for loc in env.languages:
        args = ["python", "setup.py"]
        args += [ "compile_catalog"]
        args += [ "--domain django"]
        args += [ "-d" , locale_dir ]
        args += [ "-l" , loc ]
        cmd = ' '.join(args)
        #~ must_confirm(cmd)
        local(cmd)




@task(alias='summary')
def summary(*cmdline_args):
    """Print a summary to stdout."""
    load_projects()
    headers = (
      #~ '#','Location',
      'Project','Old version','New version')

    def cells(self):
        url = self.module.SETUP_INFO['url']
        desc = "`%s <%s>`__ -- %s" % (
            self.name,url,
            self.module.SETUP_INFO['description'])
        return (
            '\n'.join(textwrap.wrap(desc,40)),
            self.dist.version,
            self.module.__version__)
        
    print rstgen.table(headers,[cells(p) for p in PROJECTS])
        
  
@task(alias='api')
def build_api(*cmdline_args):
    """
    Generate .rst files in `docs/api`.
    """
    #~ if len(env.SETUP_INFO['packages']) != 1:
        #~ abort("env.SETUP_INFO['packages'] is %s" % env.SETUP_INFO['packages'])
    
    os.environ.update(SPHINX_APIDOC_OPTIONS="members,show-inheritance")
    api_dir = env.DOCSDIR.child("api").absolute()        
    rmtree_after_confirm(api_dir)
    args = ['sphinx-apidoc']
    #~ args += ['-f'] # force the overwrite of all files that it generates.
    args += ['--no-toc'] # no modules.rst file
    args += ['-o',api_dir]
    args += [env.main_package] # packagedir 
    if False: # doesn't seem to work
        excluded = [env.ROOTDIR.child('lino','sandbox').absolute()]
        args += excluded # pathnames to be ignored
    cmd = ' '.join(args)
    #~ puts("%s> %s" % (os.getcwd(), cmd))
    #~ confirm("yes")
    local(cmd)
    

def sphinx_build_html(builder,docs_dir,cmdline_args=[],language=None):
    args = ['sphinx-build','-b',builder]
    args += cmdline_args
    #~ args += ['-a'] # all files, not only outdated
    #~ args += ['-P'] # no postmortem
    #~ args += ['-Q'] # no output
    build_dir = docs_dir.child('.build')
    build_root = docs_dir.child('.build')
    if language:
        args += ['-D', 'language=' + language] 
        args += ['-A', 'language=' + language] # needed in select_lang.html template
        build_dir = build_dir.child(language)
    if env.tolerate_sphinx_warnings:
        args += ['-w',docs_dir.child('warnings_%s.txt' % builder)]
    else:
        args += ['-W'] # consider warnings as errors
    #~ args += ['-w'+Path(env.ROOTDIR,'sphinx_doctest_warnings.txt')]
    args += [docs_dir,build_dir]
    cmd = ' '.join(args)
    local(cmd)
    
def sync_docs_data(docs_dir):
    build_dir = docs_dir.child('.build')
    for data in ('dl','data'):
        src = docs_dir.child(data).absolute()
        if src.isdir():
            target = build_dir.child('dl')
            target.mkdir()
            cmd = 'cp -ur %s %s' % (src,target.parent) 
            local(cmd)
    
    
@task(alias='userdocs')
def build_userdocs(*cmdline_args): 
    if env.languages is None: return
    docs_dir = env.ROOTDIR.child('userdocs')
    if not docs_dir.exists(): return
    for lng in env.languages:
        sphinx_build_html('html',docs_dir,cmdline_args,lng)
    dest = docs_dir.child('.build','index.html')
    docs_dir.child('index.html').copy(dest)
    sync_docs_data(docs_dir)
    
@task(alias='linkcheck')
def sphinx_build_linkcheck(*cmdline_args): 
    """sphinxbuild -b linkcheck docs."""
    docs_dir = env.ROOTDIR.child('docs')
    if docs_dir.exists(): 
        sphinx_build_html('linkcheck',docs_dir,cmdline_args)
    docs_dir = env.ROOTDIR.child('userdocs')
    if docs_dir.exists(): 
        lng = env.languages[0] # 
        sphinx_build_html('linkcheck',docs_dir,cmdline_args,lng)
    
@task(alias='docs')
def build_docs(*cmdline_args): 
    """write_readme + build sphinx html docs."""
    write_readme()
    sphinx_build_html('html',env.DOCSDIR,cmdline_args)
    sync_docs_data(env.DOCSDIR)
    
@task(alias='alldocs')
def build_all_docs(): 
    """write_readme + build ALL sphinx html docs."""
    write_readme()
    for n in ('docs','userdocs'):
        docs_dir = env.ROOTDIR.child(n)
        if docs_dir.exists(): 
            sphinx_build_html('html',docs_dir,['-a'])
            sync_docs_data(docs_dir)
    
    
    
@task(alias='clean')
def clean_html(*cmdline_args):
    """
    Delete all built Sphinx files.
    """
    build_dir = env.DOCSDIR.child('.build')
    rmtree_after_confirm(build_dir)
    if env.languages:
        build_dir = env.ROOTDIR.child('userdocs','.build')
        rmtree_after_confirm(build_dir)
    
    
#~ @task(alias='pub')
#~ def publish_all():
    #~ """
    #~ Run `publish_docs` followed by `hg_push`.
    #~ """
    #~ publish_docs()
    #~ hg_push()
    
    
@task(alias='prep')
def prepare():
    """
    Sames as `fab test html`.
    """
    run_tests()
    build_docs()
    
    
    
  
@task(alias='pub')
def publish():
    """
    Upload docs to public web server.
    """
    build_dir = env.DOCSDIR.child('.build')
    dest_url = env.docs_rsync_dest + ':~/public_html/' + env.project_name
    publish_docs(build_dir,dest_url)
    
    build_dir = env.ROOTDIR.child('userdocs','.build')
    dest_url = env.docs_rsync_dest + ':~/public_html/' + env.project_name + '-userdocs'
    if build_dir.exists():
        publish_docs(build_dir,dest_url)
    #~ if env.languages:
        #~ for lang in env.languages:
    

def publish_docs(build_dir,dest_url):
    #~ from fabric.context_managers import cd
    #~ cwd = Path(os.getcwd())
    #~ if language:
        #~ dest_url += '-userdocs/' + language
        #~ build_dir = build_dir.child(language)
        
    with lcd(build_dir):
        #~ env.BUILDDIR.chdir()
        #~ with cd(env.BUILDDIR):
        #~ addr = env.user+'@'+REMOTE.lf
        args = ['rsync','-r']
        args += ['--verbose'] 
        args += ['--progress'] # show progress
        args += ['--delete'] # delete files in dest
        args += ['--times'] # preserve timestamps
        args += ['--exclude','.doctrees'] 
        args += ['./'] # source
        args += [ dest_url ] # dest
        cmd = ' '.join(args)
        #~ must_confirm("%s> %s" % (build_dir, cmd))
        #~ confirm("yes")
        local(cmd)
        #~ cwd.chdir()
    #~ return subprocess.call(args)

#~ def run_in_demo_database(admin_cmd,*more):
	#~ if not env.demo_database: return
	#~ args = ["django-admin.py"] 
	#~ args += [admin_cmd]
	#~ args += more
	#~ args += ["--settings=" + env.demo_database]
	#~ cmd = " ".join(args)
	#~ local(cmd)
            
def run_in_demo_databases(admin_cmd,*more):
    for db in env.demo_databases:
		args = ["django-admin.py"] 
		args += [admin_cmd]
		args += more
		#~ args += ["--noinput"]
		args += ["--settings=" + db]
		#~ args += [" --pythonpath=%s" % p.absolute()]
		cmd = " ".join(args)
		local(cmd)
		
		
        #~ p = env.ROOTDIR.child(*db.split('/'))
        #~ with lcd(p):
            #~ args = ["python manage.py"] 
            #~ args += [admin_cmd]
            #~ args += more
            #~ args += [" --pythonpath=%s" % p.absolute()]
            #~ cmd = " ".join(args)
            #~ local(cmd)
  
@task()
def clean_cache():
    for db in env.demo_databases:
		m = __import__(db)
		p = Path(m.SITE.project_dir).child('media','cache')
		rmtree_after_confirm(p)
    
	
@task(alias="initdb")
def initdb_demo():
    """
    Run initdb_demo on each demo database of this project (env.demo_databases)
    """
    #~ for db in env.django_databases:
        #~ args = ["django-admin"] 
        #~ args += ["initdb_demo --settings=%s" % prj]
        #~ args += [" --pythonpath=%s" % env.DOCSDIR]
        #~ cmd = " ".join(args)
        #~ local(cmd)
        
    #~ cwd = Path(os.getcwd())
    #~ for db in env.django_databases:
        #~ env.ROOTDIR.child(db).chdir()
        #~ # cmd = 'python manage.py initdb --noinput'
        #~ args = ["django-admin"] 
        #~ args += ["initdb_demo --settings=settings"]
        #~ args += [" --pythonpath=."]
        #~ cmd = " ".join(args)
        
        #~ local(cmd)
    #~ cwd.chdir()
    
    run_in_demo_databases('initdb_demo',"--noinput")
    #~ run_in_demo_database('initdb_demo',"--noinput")
    
   

#~ @task()
#~ def runserver():
    #~ run_in_demo_databases('runserver')
    
        
@task()
def unused_run_sphinx_doctest():
    """
    Run Sphinx doctest tests. 
    Not maintained because i cannot prevent it from also trying to test 
    the documents in `django_doctests` which must be tested separately.
    """
    #~ clean_sys_path()
    #~ if sys.path[0] == '':
        #~ del sys.path[0]
    #~ print sys.path    
    #~ if len(sys.argv) > 1:
        #~ raise Exception("Unexpected command-line arguments %s" % sys.argv)
    onlythis = None
    #~ onlythis = 'docs/tutorials/human/index.rst'
    args = ['sphinx-build','-b','doctest']
    args += ['-a'] # all files, not only outdated
    args += ['-Q'] # no output
    if not onlythis:
        args += ['-W'] # consider warnings as errors
    build_dir = env.DOCSDIR.child('.build')
    args += [env.DOCSDIR,build_dir]
    if onlythis: # test only this document
        args += [ onlythis ] 
    #~ args = ['sphinx-build','-b','doctest',env.DOCSDIR,env.BUILDDIR]
    #~ raise Exception(' '.join(args))
    #~ env.DOCSDIR.chdir()
    #~ import os
    #~ print os.getcwd()
    exitcode = sphinx.main(args)
    if exitcode != 0:
        output = Path(build_dir,'output.txt')
        #~ if not output.exists():
            #~ abort("Oops: no file %s" % output)
        # six.print_("arguments to spxhinx.main() were",args)
        abort("""
=======================================
Sphinx doctest failed with exit code %s
=======================================
%s""" % (exitcode,output.read_file()))
    
@task(alias='sdist')
def setup_sdist():
    """
    Write source distribution archive file.
    """
    #~ pipy_register()
    #~ puts(env.sdist_dir)
    args = ["python", "setup.py"]
    args += [ "sdist", "--formats=gztar" ]
    args += ["--dist-dir", env.sdist_dir.child(env.SETUP_INFO['name'])]
    local(' '.join(args))
    #~ run_setup('setup.py',args)
  
@task(alias='upload')
def pypi_upload():
    """
    Upload sourcxe distribution to PyPI.
    """
    pipy_register()
    args = ["python", "setup.py"]
    args += ["sdist", "--formats=gztar" ]
    args += ["--dist-dir", env.sdist_dir.child(env.SETUP_INFO['name'])]
    args += ["upload"]
    local(' '.join(args))
    #~ run_setup('setup.py',args)
  
#~ @task()
#~ def check_packages():
    #~ """
    #~ Checks whether the `packages` list seems correct.
    #~ """
    #~ packages = find_packages()
    #~ if packages == env.SETUP_INFO['packages']: 
        #~ puts("%d packages okay" % len(packages))
        #~ return
    
    
  
@task(alias='reg')
def pipy_register():
    """
    Register to PyPI.
    """
    args = ["python", "setup.py"]
    args += ["register"]
    #~ run_setup('setup.py',args)
    local(' '.join(args))

def get_blog_entry(today):
    """
    Return an RstFile object representing the blog entry for that date.
    """
    local_root = env.work_root.child(env.blogger_project)
    parts = ('docs','blog',str(today.year),today.strftime("%m%d.rst"))
    #~ return blogdir.child(*parts)
    return RstFile(local_root,
      "http://code.google.com/p/%s/source/browse/" % env.blogger_project,
      parts)
  

@task(alias='blog')
def edit_blog_entry():
    """
    Edit today's blog entry, create an empty file if it doesn't yet exist.
    """
    today = datetime.date.today()
    entry = get_blog_entry(today)
    if not entry.path.exists():
        if confirm("Create file %s?" % entry.path):
            txt = rstgen.header(1,today.strftime(env.long_date_format))
            entry.path.write_file(txt)
            entry.path.parent.child('index.rst').set_times() # touch it for Sphinx
    args = [ os.environ['EDITOR'] ]
    args += [entry.path]
    local(' '.join(args))
  
@task(alias='ci')
def checkin():
    """
    Checkin & push to repository, using today's blog entry as commit message.
    """
    entry = get_blog_entry(datetime.date.today())
    #~ entry = Path(env.ROOTDIR,'..',env.blogger_project,*parts)
    #~ print env.ROOTDIR.parent.absolute()
    if not entry.path.exists():
        abort("%s does not exist!" % entry.path.absolute())
    #~ puts("Commit message refers to %s" % entry.absolute())
    args = ["hg","st"]
    local(' '.join(args))
    if not confirm("OK to checkin %s ?" % env.SETUP_INFO['name']):
        return 
    args = ["hg","ci"]
    args += ['-m', entry.url ]
    cmd = ' '.join(args)
    #~ confirm(cmd)
    local(cmd)
    local("hg push %s" % env.project_name)
    
#~ @task()
def unused_write_release_notes():
    """
    Generate docs/releases/x.y.z.rst file from setup_info.
    """
    v = env.SETUP_INFO['version']
    if v.endswith('+'):
        return
    notes = Path(env.ROOTDIR,'docs','releases','%s.rst' % v)
    if notes.exists():
        return
    must_confirm("Create %s" % notes.absolute())
    #~ context = dict(date=datetime.date.today().strftime(env.long_date_format))
    context = dict(date=datetime.date.today().strftime('%Y%m%d'))
    context.update(env.SETUP_INFO)
    txt = """\
==========================
Version %(version)s
==========================

Release process started :blogref:`%(date)s`


List of changes
===============

New features
------------

Optimizations
-------------

Bugfixes
--------

Manual tasks after upgrade
--------------------------


""" % context
    notes.write_file(txt)
    notes.parent.child('index.rst').set_times()
    args = [ os.environ['EDITOR'] ]
    args += [notes.absolute()]
    local(' '.join(args))
    
@task()
def write_readme():
    """
    Generate README.txt file from setup_info (if necessary).
    """
    if not env.main_package: return
    readme = env.ROOTDIR.child('README.txt')
    txt = """\
==========================
%(name)s README
==========================

%(description)s

Description
-----------

%(long_description)s

Read more on %(url)s
""" % env.SETUP_INFO
    if readme.exists() and readme.read_file() == txt:
        return 
    must_confirm("Overwrite %s" % readme.absolute())
    readme.write_file(txt)
    env.DOCSDIR.child('index.rst').set_times()
    #~ cmd = "touch " + env.DOCSDIR.child('index.rst')
    #~ local(cmd)
    #~ pipy_register()
    
#~ @task(alias='t2')
def unused_run_django_admin_tests():
    """
    Run `django-admin test` for each `env.django_admin_tests`.
    """
    for prj in env.django_admin_tests:
        cmd = "django-admin.py test --settings=%s --verbosity=0 --failfast --traceback" % prj
        local(cmd)
  
#~ @task(alias='t3')
def unused_run_django_doctests():
    """
    Run `django-admin test` in the `docs` dir for each `env.django_doctests`
    """
    #~ must_exist(env.DOCSDIR.child('manage.py'))
    #~ env.DOCSDIR.chdir()
    for prj in env.django_doctests:
        args = ["django-admin.py"] 
        args += ["test --settings=%s --failfast" % prj]
        args += [" --verbosity=0"]
        args += [" --pythonpath=%s" % env.DOCSDIR]
        cmd = " ".join(args)
        #~ cmd = "manage.py test --settings=%s --failfast" % prj
        #~ cmd = "python runtest.py %s" % prj
        local(cmd)

#~ @task(alias='t1')
def unused_run_simple_doctests():
    """
    Run a simple doctest for files specified in `env.simple_doctests`.
    """
    os.environ['DJANGO_SETTINGS_MODULE']='lino.projects.std.settings'
    for filename in env.simple_doctests:
        cmd = "python -m doctest %s" % filename
        #~ print cmd
        local(cmd)
        #~ doctest.testfile(filename, module_relative=False,encoding='utf-8')

#~ @task(alias='t4')
def unused_run_docs_doctests():
    """
    Run a simple doctest for files specified in `env.docs_doctests`.
    """
    os.environ['DJANGO_SETTINGS_MODULE']='settings'
    #~ with lcd(env.DOCSDIR):
    env.DOCSDIR.chdir()
    sys.path.insert(0,'.')
    for filename in env.docs_doctests:
        #~ cmd = "python -m doctest %s" % filename
        #~ print cmd
        #~ local(cmd)
        #~ fn = env.DOCSDIR.child(*filename.split('/'))
        doctest.testfile(filename, module_relative=False,encoding='utf-8')
    del sys.path[0]

#~ @task(alias='t5')
def unused_run_django_databases_tests():    
    """
    Run "manage.py test" for each `env.django_databases`.
    """
    run_in_demo_databases('test',"--noinput",'--failfast')

#~ @task(alias='t6')
def unused_run_setup_tests():    
    """
    Run hardcoded tests related to packaging.
    """
    unittest.main(argv=['fab','SetupTest'],module=__name__,exit=False)

#~ @task(alias='t7')
def unused_run_bash_tests():
    """
    Run the commands in `env.bash_tests`.
    """
    for cmd in env.bash_tests:
        local(cmd)
  
@task(alias='test')
def run_tests():
    """
    Run all tests
    """
    local('python setup.py -q test')
    #~ run_django_admin_tests() # t2
    #~ run_django_doctests() # t3
    #~ run_simple_doctests() # t4
    #~ run_django_databases_tests() # t5
    #~ run_setup_tests() # t6
    #~ run_bash_tests() # t7
    #~ print "run_tests() done"
    

#~ @task(alias='listpkg')
#~ def list_subpackages():
    #~ # lst = list(env.ROOTDIR.walk("__init__.py"))
    #~ for fn in env.ROOTDIR.child('lino').walk('*.py'):
        #~ print fn

@task(alias='cov')
def run_tests_coverage():
    """
    Run all tests, creating coverage report
    """
    import coverage
    #~ clean_sys_path()
    puts("Running tests for '%s' within coverage..." % env.project_name)
    #~ env.DOCSDIR.chdir()
    source = []
    for package_name in env.SETUP_INFO['packages']:
        m = __import__(package_name)
        source.append(os.path.dirname(m.__file__))
    #~ cov = coverage.coverage(source=['djangosite'])
    if not confirm("coverage source=%s" % source):
        abort
    cov = coverage.coverage(source=source)
    #~ cov = coverage.coverage()
    cov.start()

    # .. call your code ..
    rv = run_tests()

    cov.stop()
    cov.save()

    cov.html_report()    
    return rv
    




# -*- coding: utf-8 -*-
r"""
A suite of utilities to programmatically generate rst source text.

:copyright: Copyright 2011-2013 by Luc Saffre.
:license: BSD, see LICENSE for more details.

Especially the :func:`table` function is used by the 
:class:`complextable <djangosite.utils.sphinxconf.ComplexTableDirective>` 
directive and by 
:meth:`Table.to_rst <lino.core.actions.ActionRequest.to_rst>`.
Here we present the raw API.

Usage examples
--------------

Here is the data we are going to render into different tables:

>>> headers = ["Country","City","Name"]
>>> rows = []
>>> rows.append(["Belgium","Eupen","Gerd"])
>>> rows.append(["Estonia","Vigala","Luc"])
>>> rows.append(["St. Vincent and the Grenadines","Chateaubelair","Nicole"])


The simplest case of :func:`table`:

.. complextable::
  :header: 
  
  Code <NEXTCELL> Result <NEXTROW>

  >>> from atelier.rstgen import table
  >>> print table(headers,rows)
  ================================ =============== ========
   Country                          City            Name
  -------------------------------- --------------- --------
   Belgium                          Eupen           Gerd
   Estonia                          Vigala          Luc
   St. Vincent and the Grenadines   Chateaubelair   Nicole
  ================================ =============== ========
  <BLANKLINE>
  
  <NEXTCELL>
  
  `\ ` 
  
  ================================ =============== ========
   Country                          City            Name
  -------------------------------- --------------- --------
   Belgium                          Eupen           Gerd
   Estonia                          Vigala          Luc
   St. Vincent and the Grenadines   Chateaubelair   Nicole
  ================================ =============== ========

A table without headers:

.. complextable::
  :header: 
  
  Code <NEXTCELL> Result <NEXTROW>

  >>> print table(headers,rows,show_headers=False)
  ================================ =============== ========
   Belgium                          Eupen           Gerd
   Estonia                          Vigala          Luc
   St. Vincent and the Grenadines   Chateaubelair   Nicole
  ================================ =============== ========
  <BLANKLINE>

  <NEXTCELL>
  
  `\ ` 
  
  ================================ =============== ========
   Belgium                          Eupen           Gerd
   Estonia                          Vigala          Luc
   St. Vincent and the Grenadines   Chateaubelair   Nicole
  ================================ =============== ========
  
You might prefer to use directly the :class:`Table` class:

.. complextable::
  :header: 
  
  Code <NEXTCELL> Result <NEXTROW>

  >>> from atelier.rstgen import Table
  >>> t = Table(headers)
  >>> print t.to_rst(rows)
  ================================ =============== ========
   Country                          City            Name
  -------------------------------- --------------- --------
   Belgium                          Eupen           Gerd
   Estonia                          Vigala          Luc
   St. Vincent and the Grenadines   Chateaubelair   Nicole
  ================================ =============== ========
  <BLANKLINE>

  <NEXTCELL>
  
  `\ ` 
  
  ================================ =============== ========
   Country                          City            Name
  -------------------------------- --------------- --------
   Belgium                          Eupen           Gerd
   Estonia                          Vigala          Luc
   St. Vincent and the Grenadines   Chateaubelair   Nicole
  ================================ =============== ========

If there is at least one cell that contains a newline character,
the result will be a complex table:

.. complextable::
  :header: 
  
  Code <NEXTCELL> Result <NEXTROW>

  >>> rows[2] = ['''St. Vincent
  ... and the Grenadines''',"Chateaubelair","Nicole"]
  >>> print table(headers,rows)
  +--------------------+---------------+--------+
  | Country            | City          | Name   |
  +====================+===============+========+
  | Belgium            | Eupen         | Gerd   |
  +--------------------+---------------+--------+
  | Estonia            | Vigala        | Luc    |
  +--------------------+---------------+--------+
  | St. Vincent        | Chateaubelair | Nicole |
  | and the Grenadines |               |        |
  +--------------------+---------------+--------+
  <BLANKLINE>

  <NEXTCELL>
  
  `\ ` 

  +--------------------+---------------+--------+
  | Country            | City          | Name   |
  +====================+===============+========+
  | Belgium            | Eupen         | Gerd   |
  +--------------------+---------------+--------+
  | Estonia            | Vigala        | Luc    |
  +--------------------+---------------+--------+
  | St. Vincent        | Chateaubelair | Nicole |
  | and the Grenadines |               |        |
  +--------------------+---------------+--------+
  
"""

#~ import cStringIO as StringIO
import StringIO

class Column(object):
    def __init__(self,index,header,width=None):
        self.header = header
        self.width = width
        self.index = index
        
    def adjust_width(self,row):
        s = unicode(row[self.index])
        for ln in s.splitlines():
            if self.width is None or self.width < len(ln):
                self.width = len(ln)

def write_header(fd,level,s):
    def writeln(s=''):
        fd.write(s+'\n')
    _write_header(writeln,level,s)
    
def header(level,text):
    """
    Render the text as a header with the specified level.
    """
    result = StringIO.StringIO()
    def writeln(s=''):
        result.write(s + '\n')
    _write_header(writeln,level,text)
    return result.getvalue()
        
def _write_header(writeln,level,s):
    if level == 1:
        writeln('=' * len(s))
    elif level == 2:
        writeln('-' * len(s))
    writeln(s)
    if level == 1:
        writeln('=' * len(s))
    elif level == 2:
        writeln('-' * len(s))
    elif level == 3:
        writeln('=' * len(s))
    elif level == 4:
        writeln('-' * len(s))
    else:
        raise Exception("Invalid level %d" % level)
    writeln()
    
class Table(object):
    """
    Renders as a table.
    
    """
    simple = True
    
    def convert(self,v):
        return unicode(v)
    
    def __init__(self,headers,show_headers=True):
        self.headers = [self.convert(h) for h in headers]
        self.show_headers = show_headers
        self.cols = [ Column(i,h) for i,h in enumerate(headers)]
        self.adjust_widths(headers)
          
        
    def adjust_widths(self,row):
        if len(row) != len(self.headers):
            raise Exception("Invalid row %(row)s" % dict(row=row))
        for col in self.cols:
            col.adjust_width(row)
            if '\n' in row[col.index]:
                self.simple = False
      
    def format_row(self,row):
        #~ return ' '.join([unicode(row[c.index]).ljust(c.width) for c in self.cols])
        lines = [ [] for x in self.cols ]
        for c in self.cols:
            cell = row[c.index]
            for ln in cell.splitlines():
                lines[c.index].append(ln.ljust(c.width))
        height = 1
        for c in self.cols:
            height = max(height,len(lines[c.index]))
        for c in self.cols:
            while len(lines[c.index]) < height:
                lines[c.index].append(''.ljust(c.width))
        x = []
        for i in range(height):
            x.append(self.margin 
                + self.colsep.join([' '+lines[c.index][i]+' ' for c in self.cols]) 
                + self.margin)
        return '\n'.join(x)
        
    def write(self,fd,data):
        if len(data) == 0:
            raise Exception("No rows in %r" % data)
        rows = []
        for i,row in enumerate(data):
            assert len(row) == len(self.cols)
            rows.append([self.convert(v) for v in row])
              
        for row in rows: self.adjust_widths(row)
          
        if self.simple:
            self.header1 = ' '.join([('=' * (c.width+2)) for c in self.cols])
            self.header2 = ' '.join([('-' * (c.width+2)) for c in self.cols])
            self.margin = '' 
            self.colsep = ' ' 
        else:
            self.header1 = '+'+'+'.join([('-' * (c.width+2)) for c in self.cols])+'+'
            self.header2 = '+'+'+'.join([('=' * (c.width+2)) for c in self.cols])+'+'
            self.margin = '|' 
            self.colsep = '|' 
            
        def writeln(s):
            fd.write(s.rstrip()+'\n')
            
            
        writeln(self.header1)
        if self.show_headers:
            writeln(self.format_row(self.headers))
            writeln(self.header2)
        for row in rows:
            writeln(self.format_row(row))
            if not self.simple:
                writeln(self.header1)
        if self.simple:
            writeln(self.header1)
          
    def to_rst(self,rows):
        fd = StringIO.StringIO()
        self.write(fd,rows)
        return fd.getvalue()
        
          


def table(headers,rows,**kw):
    t = Table(headers,**kw)
    return t.to_rst(rows)
    
    
#~ def py2rst(v):
    #~ from django.db import models 
    #~ if issubclass(v,models.Model):
        #~ headers = ("name","verbose name","type","help text")
        #~ rows = [
          #~ (f.name,f.verbose_name,f.__class__.__name__,f.help_text)
          #~ for f in v._meta.fields
        #~ ]
        #~ return table(headers,rows)
    #~ return unicode(v)
    
    
    
def ul(items):
    r"""
    >>> print ul([
    ...   "A first item\nwith several lines of text.",
    ...   "Another item with a nested paragraph:\n\n  Like this.\n\nWow."])
    <BLANKLINE>
    - A first item
      with several lines of text.
    <BLANKLINE>
    - Another item with a nested paragraph:
    <BLANKLINE>
        Like this.
    <BLANKLINE>
      Wow.
    <BLANKLINE>

    """
    s = ""
    for i in items:
        text = '\n  '.join(i.splitlines())
        s += "\n- %s\n" % text
    return s
    
    
    
def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()

    
